//============================================================================
// Name        : cp2dPackingRotations.cpp
// Author      : Pedro Henrique Del Bianco Hokama
// Version     :
// Copyright   : Free for academics
// Description : This lib solves the packing problem with
//               rotations, using the cpoptimizer
//============================================================================


#include <vector>
#include <algorithm>
#include <limits>
#include "ilcp/cp.h"
#include "cp2dPackingRotations.h"


using namespace std;

bool cp_2opp_simple_unloading(int W, int H, int n,
    int *w, int *h, int *p,
    int *x, int *y, double timeLimit, double *timeUsed,
    int *status, int verbose){
  IloEnv env;
  bool feasSol = false;

  try {
    IloModel model(env);
    //Model the data
    IloInt nbItens = n;
    IloInt iloW = W;
    IloInt iloH = H;

    //Variables of the problem
    IloIntVarArray X(env);
    IloIntVarArray Y(env);
    IloIntVarArray ilow(env);
    IloIntVarArray iloh(env);

    //for each item
    for (int i = 0; i < n; i++) {
      // create each position variable with domain in x_i = 0 .. W; y_i = 0 .. H
      ostringstream vname;
      vname << "x_" << i;
      X.add(IloIntVar(env, 0, iloW, vname.str().c_str()));
      vname.str("");
      vname << "y_" << i;
      Y.add(IloIntVar(env, 0, iloH, vname.str().c_str()));

      ostringstream wname; wname.str("");
      wname << "w_" << i;

      ostringstream hname; hname.str("");
      hname << "h_" << i;

      if(w[i] != h[i]){
        IloIntArray ilo_feas_dimension(env, 2);
        ilo_feas_dimension[0] = w[i];
        ilo_feas_dimension[1] = h[i];
        ilow.add(IloIntVar(env, ilo_feas_dimension, wname.str().c_str()));
        iloh.add(IloIntVar(env, ilo_feas_dimension, hname.str().c_str()));
        model.add(ilow[i] != iloh[i]);
      }else{
        ilow.add(IloIntVar(env, w[i], w[i]));
        iloh.add(IloIntVar(env, w[i], w[i]));
      }
    }

    //overlap constraint
    for (int i = 0; i < nbItens; i++) {
      model.add(X[i] + ilow[i] <= iloW);
      model.add(Y[i] + iloh[i] <= iloH);
      for (int j = i + 1; j < nbItens; j++) {
        if (p[i] < p[j]){
          model.add((X[i] + ilow[i] <= X[j])                              || (Y[j] + iloh[j] <= Y[i]) || (X[j] + ilow[j] <= X[i]));
        }else if(p[j] < p[i]){
          model.add((X[i] + ilow[i] <= X[j]) || (Y[i] + iloh[i] <= Y[j])                              || (X[j] + ilow[j] <= X[i]));
        }else{
          model.add((X[i] + ilow[i] <= X[j]) || (Y[i] + iloh[i] <= Y[j])  || (Y[j] + iloh[j] <= Y[i]) || (X[j] + ilow[j] <= X[i]));
        }
      }
    }

    //Create an instance of IloCP
    IloCP cp(model);

    if(verbose < 3) cp.setParameter(IloCP::PropagationLog, IloCP::Quiet);
    if(verbose < 3) cp.setParameter(IloCP::LogVerbosity, IloCP::Quiet);
    if(timeLimit > 0) cp.setParameter(IloCP::TimeLimit, timeLimit);

    if(verbose > 2) cout << "Starting CP Solver" << endl;
    cp.exportModel("model.cpo");

    //Search for a solution
    if (cp.solve()) {
      for (int j = 0; j < nbItens; j++) {
        x[j] = cp.getValue(X[j]);
        y[j] = cp.getValue(Y[j]);
        w[j] = cp.getValue(ilow[j]);
        h[j] = cp.getValue(iloh[j]);
      }
      feasSol = true;

      *timeUsed = cp.getInfo(cp.SolveTime);
      *status = (int) cp.getStatus();
    } else {
      *timeUsed = cp.getInfo(cp.SolveTime);
      *status = (int) cp.getStatus();
    }
    if(verbose > 2) cout << "Finishing CP Solver" << endl;
    cp.end();
  }
  catch(IloException & e) {
    cout << "Erro" << endl;
    env.out() << " ERROR: " << e.getMessage() << std::endl;
  }

  env.end();
  return feasSol;
}

