/*
 * irp_subtourelim.h
 *
 *  Created on: 17 de out de 2016
 *      Author: hokama
 */

#ifndef IRP2D_SUBTOURELIM2D_H_
#define IRP2D_SUBTOURELIM2D_H_

#define MAX_NUMBER_OF_ITEMS_ON_ROUTE 10000

#include "gurobi_c++.h"

#include <sstream>
#include <iostream>
#include <iomanip>

#include "irp2d_instance.h"
#include "cp2dPackingRotations.h"

namespace IRP2D {
using namespace std;

enum Route_feasibility { FEASIBLE, INFEASIBLE_DISCONNECTED, INFEASIBLE_PACKING};

int getPacking(Irp2dInstance *inst, vector < pair < int, int > > route);

int getPacking(Irp2dInstance *inst, vector < int >  route, int period);

void printPackings(Irp2dInstance *inst);

class subtourelim2d: public GRBCallback{

public:
  vector < vector < GRBVar * > > x;
  vector < vector < GRBVar * > > y;
  vector < vector < vector < GRBVar * > > > q;
  double time_for_packing_cp_solver;
  Irp2dInstance *inst;
  int verbose;

  //Constructors
  subtourelim2d(vector < vector < GRBVar * > > xp, vector < vector < GRBVar * > > yp, Irp2dInstance *instp, int verbosep, double time_limit) {
    x = xp;
    y = yp;
    inst = instp;
    verbose = verbosep;
    q.clear();
    time_for_packing_cp_solver = time_limit;
  }

  subtourelim2d(vector < vector < GRBVar * > > xp,
      vector < vector < GRBVar * > > yp,
      vector < vector < vector < GRBVar * > > > qp,
      Irp2dInstance *instp, int verbosep, double time_limit) {
    x = xp;
    y = yp;
    q = qp;
    inst = instp;
    verbose = verbosep;
    time_for_packing_cp_solver = time_limit;
  }

protected:
  void callback() {
    try {
      if (where == GRB_CB_MIPSOL) {
        vector < GRBTempConstr >  bad_cycles;
        if (find_subtour(x, y, inst, bad_cycles) > 0) {
          // Add subtour elimination constraint
          for (unsigned int c = 0; c < bad_cycles.size(); c++){
            if(verbose >= 3) cout << "Disconnect subtour found" << endl;
            addLazy(bad_cycles[c]);
          }
        }else if (find_tour(x, y, q, inst, bad_cycles) > 0) {
          // Add subtour elimination constraint
          if(verbose >= 3) cerr << "No subtour found - invalid cycle " << endl;
          for (unsigned int c = 0; c < bad_cycles.size(); c++){
            addLazy(bad_cycles[c]);
          }
        }
      }
    } catch (GRBException &e) {
      cout << "Error number: " << e.getErrorCode() << endl;
      cout << e.getMessage() << endl;
    } catch (...) {
      cout << "Error during callback" << endl;
    }
  }

  //Verifies if a route, and a set of demands has a feasible packing
  bool valid_packing(vector < pair < int, int > > route,
      Irp2dInstance *inst){

    int w[MAX_NUMBER_OF_ITEMS_ON_ROUTE];
    int h[MAX_NUMBER_OF_ITEMS_ON_ROUTE];
    int p[MAX_NUMBER_OF_ITEMS_ON_ROUTE];
    int number_of_itens = 0;
    int item_count = 0;

    int vertex_order = 0;
    for (unsigned i = 0; i < route.size(); i++){
      int vertex = route[i].first;
      int demand_period = route[i].second;
      //from the second order, check if visiting a new vertex
      if(i >= 1 && vertex != route[i-1].first) vertex_order++;
      if(vertex == 0) continue;
      for (int product = 0; product < inst->get_number_of_products(); product++){
        if (verbose >= 3) cout << "Demand (" << vertex << "," << demand_period << ") of product " << product << " is attended by this route " << endl;
        int items_of_product = inst->get_demand_of_product_per_customer_per_period(vertex,product,demand_period);
        for (int item = 0; item < items_of_product; item++){
          w[item_count] = inst->get_width(product);
          h[item_count] = inst->get_height(product);
          p[item_count] = vertex_order;
          item_count++;
        }
        number_of_itens += items_of_product;
      }
    }

    if(number_of_itens <= 1) return true;

    int route_id = getPacking(inst, route);

    if(route_id != -1){
      if(verbose >= 3){
        cout << "Route found on table: ";
        for(unsigned i = 0; i < route.size(); i++) cout << "(" << route[i].first << ", " << route[i].second << ") ";
        cout << endl;
      }
      return inst->feasible_routes[route_id];
    }

    //Route not found on table, check feasibility
    bool feasible;
    inst->routes.push_back(route);
    int * x_ = (int *) calloc(MAX_NUMBER_OF_ITEMS_ON_ROUTE, sizeof(int));
    int * y_ = (int *) calloc(MAX_NUMBER_OF_ITEMS_ON_ROUTE, sizeof(int));
    if(verbose >= 3){
      cout << "Total items in this cycle " << number_of_itens << endl;
      cout << "Verifying feasibility... " << endl;
      cout << "    w    h    p " << endl;
      for (int i = 0; i< number_of_itens; i++){
        cout << setw(5) << w[i]; cout << setw(5) << h[i]; cout << setw(5) << p[i]; cout << endl;
      }
    }
    int pack_res = 0; //cp2opp_unloading_constraint(number_of_itens, inst->get_vehicle_width(), inst->get_vehicle_height(), w, h, x_, y_, p, 2);

    if(pack_res!=1){
      pack_res = 0;
      double timeUsed;
      int status;
      pack_res = cp_2opp_simple_unloading(inst->get_vehicle_width(), inst->get_vehicle_height(), number_of_itens, w, h, p, x_, y_, time_for_packing_cp_solver, &timeUsed, &status, verbose);
    }
    feasible = number_of_itens > 1 ? pack_res : true;

    if(verbose >= 3) cout << "Feasibility: " << feasible << endl;
    vector < int > x_route, y_route, w_route, h_route, p_route;
    inst->feasible_routes.push_back(feasible);
    if (feasible){
      if(verbose >= 3) cout << "    x    y" << endl;
      for (int i = 0; i< number_of_itens; i++){
        if(verbose >= 3) {cout << setw(5) << x_[i]; cout << setw(5) << y_[i] << endl;}
        x_route.push_back(x_[i]);
        y_route.push_back(y_[i]);
        w_route.push_back(w[i]);
        h_route.push_back(h[i]);
        p_route.push_back(p[i]);
      }
    }
    free(x_); free(y_);
    inst->x_routes.push_back(x_route); inst->y_routes.push_back(y_route);
    inst->w_routes.push_back(w_route); inst->h_routes.push_back(h_route);
    inst->p_routes.push_back(p_route);

    return feasible;
  }

  //Verifies if a given route is valid, considering connectivity constraint
  //and packing constraint
  //Returns 1 if route is valid
  //        0 if is subtour
  Route_feasibility valid_cycle(vector < pair <int,int> > route, Irp2dInstance *inst){
    if (verbose >= 3) {
      cout << "Verifying route: ";
      for (unsigned u = 0; u < route.size(); u++){
        cout << "(" << route[u].first << ", " << route[u].second << ")" ;
      }
      cout << endl;
    }
    if (route.size() == 1) return FEASIBLE;
    for (unsigned i = 0; i < route.size(); i++){
      if (route[i].first == 0) {
        if(valid_packing(route, inst) == true){
          return FEASIBLE;
        }else return INFEASIBLE_PACKING;
      }
    }
    return INFEASIBLE_DISCONNECTED;
  }

  int valid_subtour(int *tour, int start, int end, Irp2dInstance *inst, int vehicle, int period){
    if (end-start == 1) return 1;
    for (int i = start; i < end; i++){
      if (tour[i] == 0) {
        return 1;
      }
    }
    return 0;
  }
  int find_subtour(vector < vector < GRBVar * > > x,
      vector < vector < GRBVar * > > y,
      Irp2dInstance *inst, vector < GRBTempConstr > &bad_cycles) {

    int number_of_vertices = inst->get_number_of_customers()+1;
    for(int period = 1; period < inst->get_horizon() + 1; period++){
      for(int vehicle = 0; vehicle < inst->get_number_of_vehicles(); vehicle++){

        //starting looking for cycles
        bool seen[number_of_vertices];
        int i, node, len, start;
        int tour[number_of_vertices + inst->get_number_of_vehicles()];

        for (int vertex = 0; vertex < number_of_vertices; vertex++)
          seen[vertex] = false;

        //start position in tour
        start = 0;
        node = 0;

        while(start < number_of_vertices){
          //searching for the next unseen node
          for (node = 0; node < number_of_vertices; node++)
            if (!seen[node]) break;
          //all nodes already seen
          if (node == number_of_vertices) break;

          //adding nodes to tour
          for (len = 0; len < number_of_vertices; len++) {
            //add node
            tour[start+len] = node;
            seen[node] = true;
            //search next
            for (i = 0; i < number_of_vertices; i++) {
              if (getSolution(x[inst->get_edge_id(i,node)][vehicle][period]) > 0.9 && !seen[i]) {
                //set next
                node = i;
                break;
              }
            }
            if (i == number_of_vertices) {
              //if no 'next' node were found, cycle ended
              //lets take a look in this cycle
              len++;
              int valid_cycle_ret = valid_subtour(tour, start, start+len, inst, vehicle, period);
              //If cycle is invalid
              if (valid_cycle_ret == 0) {

                GRBLinExpr cycle;
                for (int u = start; u < start + len; u++){
                  for (int v = u+1; v < start + len; v++){
                    cycle += x[  inst->get_edge_id(    tour[u],           tour[v])    ][     vehicle     ][     period];
                    if (verbose >= 2) cout << "+x(" << tour[u] << ", " << tour[v] << ")[" << vehicle << "][" << period << "]";
                  }
                  cycle -= y[                        tour[u] ][         vehicle][          period];
                  if (verbose >= 2) cout << "-y[" << tour[u] << "][" << vehicle << "][" << period << "]";
                }
                //cycle +=                    y[     start     ][     vehicle     ][     period     ];
                //if (verbose >= 2) cout << "+y[" << start << "][" << vehicle << "][" << period << "]";

                //if (verbose >= 2) cout << "<= " << 0 << endl;

                for (int u = start; u < start + len; u++){
                  bad_cycles.push_back(GRBTempConstr(cycle <= -y[     tour[u]     ][     vehicle     ][     period     ]));
                  if (verbose >= 2) cout <<                "<= -y["<< tour[u]<<  "][" << vehicle << "][" << period << "]" << endl;
                }
              }

              //start next cycle
              start += len;
              break;
            }
            //proceed to next node
          }//end of this cycle

        }//end while,. all nodes considered

      }//end of vehicle
    }//end of period

    return bad_cycles.size();
  }

  //If set is invalid
  void treat_cycle(Route_feasibility valid_cycle_ret, vector<int> route_nodes, vector< pair <int,int> > route,
      int vehicle, int period, vector < GRBTempConstr > &bad_cycles){

    if (valid_cycle_ret == INFEASIBLE_DISCONNECTED) {
      if (verbose >= 3) cout << "Subtour found"<< endl;
      GRBLinExpr cycle;
      for (unsigned u = 0; u < route_nodes.size(); u++){
        for (unsigned v = u+1; v < route_nodes.size(); v++){
          cycle += x[  inst->get_edge_id(    route_nodes[u],           route_nodes[v])    ][     vehicle     ][     period];
          if (verbose >= 3) cout << "+x(" << route_nodes[u] << ", " << route_nodes[v] << ")[" << vehicle << "][" << period << "]";
        }
        cycle -= y[                        route_nodes[u] ][         vehicle][          period];
        if (verbose >= 3) cout << "-y[" << route_nodes[u] << "][" << vehicle << "][" << period << "]";
      }
      if (verbose >= 3) cout << endl;
      for (unsigned u = 0; u < route_nodes.size(); u++){
        bad_cycles.push_back(GRBTempConstr(cycle <= -y[     route_nodes[u]     ][     vehicle     ][     period     ]));
      }
    }

    //this cycle is infeasible
    if (valid_cycle_ret == INFEASIBLE_PACKING) {
      GRBLinExpr cycle;
      int total_vars = 0;
      for (unsigned u = 0; u < route_nodes.size() - 1; u++){
        GRBVar edge = x[  inst->get_edge_id(    route_nodes[u],           route_nodes[u+1])    ][     vehicle     ][     period];
        if(getSolution(edge) >= 0.9){
          cycle += x[  inst->get_edge_id(    route_nodes[u],           route_nodes[u+1])    ][     vehicle     ][     period];
          if (verbose >= 3) cerr << "+x(" << route_nodes[u] << ", " << route_nodes[u+1] << ")[" << vehicle << "][" << period << "]";
          total_vars++;
        }
      }
      for (unsigned u = 0; u < route.size(); u++){
        if(route[u].first == 0) continue;
        for (int t = period; t <= route[u].second; t++){
          if(verbose >= 3) cerr << "q[ " << route[u].first << "][" << route[u].second << "][" << vehicle << "][" << t << "]" << endl;
          GRBVar q_var = q[route[u].first][route[u].second][vehicle][t];
          if(getSolution(q_var) >= 0.9){
            cycle += q_var;
            if (verbose >= 3) cout << q_var.get(GRB_StringAttr_VarName);
            total_vars++;
          }
        }
      }
      if(verbose >= 3) cerr << " <= " << total_vars - 1 << " infeas_packing" << endl;

      bad_cycles.push_back(GRBTempConstr(cycle <= total_vars - 1));
    }
  }

  void get_route_of_demands(vector < int > &route_nodes, vector < pair < int, int>  > &route,  vector < vector < vector < GRBVar * > > > &q,
      int vehicle, int period){

    if(verbose >= 3){
      cout << "Route found route_nodes = ";
      for (unsigned v = 0; v < route_nodes.size();v++){
        cout << route_nodes[v] << " ";
      }
      cout << endl;
    }

    for (unsigned v = 0; v < route_nodes.size();v++){
      int route_node = route_nodes[v];
      if(route_node == 0){
        pair < int, int> demand (route_node, -1);
        route.push_back(demand);
        continue;
      }
      bool has_demand = false;
      for (int t = period; t <= inst->get_horizon(); t++){
        if(getSolution(q[route_node][t][vehicle][period]) > 0.9){
          if(verbose >= 3) cout << "(" << route_nodes[v] << ", " << t << ")" ;
          pair < int, int> demand (route_node, t);
          route.push_back(demand);
          has_demand = true;
        }
      }
      if(has_demand == false){
        if(verbose >= 3) cout << "(" << route_nodes[v] << ", " << -1 << ")" ;
        pair < int, int> demand (route_node, -1);
        route.push_back(demand);
        has_demand = true;
      }
    }
    if(verbose >= 3) cerr << endl;
  }

  //Returns in bad_cycles, violated constraints considering disconnected
  //tour, and routes with infeasible packing, considering given vehicle and period.
  //It receives a vector of x[edge][vehicle][period], y[client][vehicle][period]
  //and q[client][t_demand][vehicle][period] (may be empty or null if just_in_time
  //flag is active
  void find_tour_in_vehicle_in_period(vector < vector < GRBVar * > > &x,
      vector < vector < GRBVar * > > &y,
      vector < vector < vector < GRBVar * > > > &q,
      Irp2dInstance *inst, vector < GRBTempConstr > &bad_cycles, int vehicle, int period){

    int number_of_vertices = inst->get_number_of_customers()+1;
    //starting looking for cycles
    bool seen[number_of_vertices];
    int i, node;

    for (int vertex = 0; vertex < number_of_vertices; vertex++)
      seen[vertex] = false;

    //start position in tour
    node = 0;

    //while(start < number_of_vertices){
    while(true){
      //searching for the next unseen node
      for (node = 0; node < number_of_vertices; node++)
        if (!seen[node]) break;
      //all nodes already seen
      if (node == number_of_vertices) break;
      //route[r].first = node visited
      //route[r].second = t_demand of node attended (-1 if none)
      vector < pair < int, int>  > route;
      vector < int > route_nodes;

      //adding nodes to tour
      //for (len = 0; len < number_of_vertices; len++) {
      while(true){
        //add node
        route_nodes.push_back(node);
        seen[node] = true;
        //search next
        for (i = 0; i < number_of_vertices; i++) {
          if (!seen[i] && getSolution(x[inst->get_edge_id(i,node)][vehicle][period]) > 0.9) {
            //set next
            node = i;
            break;
          }
        }
        if (i == number_of_vertices) {
          //if no 'next' node were found, cycle ended
          //lets take a look in this cycle
          //len++;
          get_route_of_demands(route_nodes, route, q, vehicle, period);
          Route_feasibility valid_cycle_ret = valid_cycle(route, inst);

          treat_cycle(valid_cycle_ret, route_nodes, route, vehicle, period, bad_cycles);
          break;
        }
        //proceed to next node
      }//end of this cycle

    }//end for,. all nodes considered

  }

  //Returns in bad_cycles, violated constraints considering disconnected
  //tour, and routes with infeasible packing.
  //It receives a vector of x[edge][vehicle][period], y[client][vehicle][period]
  //and q[client][t_demand][vehicle][period] (may be empty or null if just_in_time
  //flag is active
  int find_tour(vector < vector < GRBVar * > > &x,
      vector < vector < GRBVar * > > &y,
      vector < vector < vector < GRBVar * > > > &q,
      Irp2dInstance *inst, vector < GRBTempConstr > &bad_cycles) {

    for(int period = 1; period <= inst->get_horizon(); period++){
      for(int vehicle = 0; vehicle < inst->get_number_of_vehicles(); vehicle++){
        if (verbose >= 3) cout << "Verifying vehicle " << vehicle << " in period " << period << endl;
        find_tour_in_vehicle_in_period(x, y, q, inst, bad_cycles, vehicle, period);

      }//end of vehicle
    }//end of period

    return bad_cycles.size();
  }

};

}// end of namespace

#endif /* IRP2D_SUBTOURELIM2D_H_ */
